# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_device": "00_setup.ipynb",
         "read_settings": "00_setup.ipynb",
         "DEVICE": "00_setup.ipynb",
         "settings": "00_setup.ipynb",
         "DATA_STORE": "00_setup.ipynb",
         "LOG_STORE": "00_setup.ipynb",
         "MODEL_STORE": "00_setup.ipynb",
         "EXPERIMENT_STORE": "00_setup.ipynb",
         "PATH_1K": "00_setup.ipynb",
         "PATH_10K": "00_setup.ipynb",
         "PATH_20K": "00_setup.ipynb",
         "PATH_100K": "00_setup.ipynb",
         "FILENAMES": "00_setup.ipynb",
         "SYNTHEA_DATAGEN_DATES": "00_setup.ipynb",
         "CONDITIONS": "00_setup.ipynb",
         "LABELS": "00_setup.ipynb",
         "LOG_NUMERICALIZE_EXCEP": "00_setup.ipynb",
         "read_raw_ehrdata": "01_preprocessing_clean.ipynb",
         "split_patients": "01_preprocessing_clean.ipynb",
         "split_ehr_dataset": "01_preprocessing_clean.ipynb",
         "cleanup_pts": "01_preprocessing_clean.ipynb",
         "cleanup_obs": "01_preprocessing_clean.ipynb",
         "cleanup_algs": "01_preprocessing_clean.ipynb",
         "cleanup_crpls": "01_preprocessing_clean.ipynb",
         "cleanup_meds": "01_preprocessing_clean.ipynb",
         "cleanup_img": "01_preprocessing_clean.ipynb",
         "cleanup_procs": "01_preprocessing_clean.ipynb",
         "cleanup_cnds": "01_preprocessing_clean.ipynb",
         "cleanup_immns": "01_preprocessing_clean.ipynb",
         "cleanup_dataset": "01_preprocessing_clean.ipynb",
         "extract_ys": "01_preprocessing_clean.ipynb",
         "insert_age": "01_preprocessing_clean.ipynb",
         "clean_raw_ehrdata": "01_preprocessing_clean.ipynb",
         "load_cleaned_ehrdata": "01_preprocessing_clean.ipynb",
         "load_ehr_vocabcodes": "01_preprocessing_clean.ipynb",
         "EhrVocab": "02_preprocessing_vocab.ipynb",
         "ObsVocab": "02_preprocessing_vocab.ipynb",
         "EhrVocabList": "02_preprocessing_vocab.ipynb",
         "get_all_emb_dims": "02_preprocessing_vocab.ipynb",
         "collate_codes_offsts": "03_preprocessing_transform.ipynb",
         "get_codenums_offsts": "03_preprocessing_transform.ipynb",
         "get_demographics": "03_preprocessing_transform.ipynb",
         "Patient": "03_preprocessing_transform.ipynb",
         "get_pckl_dir": "03_preprocessing_transform.ipynb",
         "PatientList": "03_preprocessing_transform.ipynb",
         "cpu_cnt": "03_preprocessing_transform.ipynb",
         "create_all_ptlists": "03_preprocessing_transform.ipynb",
         "preprocess_ehr_dataset": "03_preprocessing_transform.ipynb",
         "EHRDataSplits": "04_data.ipynb",
         "LabelEHRData": "04_data.ipynb",
         "EHRDataset": "04_data.ipynb",
         "EHRData": "04_data.ipynb",
         "accuracy": "05_metrics.ipynb",
         "null_accuracy": "05_metrics.ipynb",
         "ROC": "05_metrics.ipynb",
         "MultiLabelROC": "05_metrics.ipynb",
         "plot_rocs": "05_metrics.ipynb",
         "plot_train_valid_rocs": "05_metrics.ipynb",
         "auroc_score": "05_metrics.ipynb",
         "auroc_ci": "05_metrics.ipynb",
         "save_to_checkpoint": "06_learn.ipynb",
         "load_from_checkpoint": "06_learn.ipynb",
         "get_loss_fn": "06_learn.ipynb",
         "RunHistory": "06_learn.ipynb",
         "train": "06_learn.ipynb",
         "evaluate": "06_learn.ipynb",
         "fit": "06_learn.ipynb",
         "predict": "06_learn.ipynb",
         "plot_loss": "06_learn.ipynb",
         "plot_losses": "06_learn.ipynb",
         "plot_aurocs": "06_learn.ipynb",
         "plot_train_valid_aurocs": "06_learn.ipynb",
         "plot_fit_results": "06_learn.ipynb",
         "summarize_prediction": "06_learn.ipynb",
         "count_parameters": "06_learn.ipynb",
         "dropout_mask": "07_models.ipynb",
         "InputDropout": "07_models.ipynb",
         "linear_layer": "07_models.ipynb",
         "create_linear_layers": "07_models.ipynb",
         "init_lstm": "07_models.ipynb",
         "EHR_LSTM": "07_models.ipynb",
         "init_cnn": "07_models.ipynb",
         "conv_layer": "07_models.ipynb",
         "EHR_CNN": "07_models.ipynb",
         "get_data": "08_experiment.ipynb",
         "get_optimizer": "08_experiment.ipynb",
         "get_model": "08_experiment.ipynb",
         "Experiment": "08_experiment.ipynb"}

modules = ["setup.py",
           "preprocessing/clean.py",
           "preprocessing/vocab.py",
           "preprocessing/transform.py",
           "data.py",
           "metrics.py",
           "learn.py",
           "models.py",
           "experiment.py"]

doc_url = "https://corazonlabs.github.io/lemonade/"

git_url = "https://github.com/corazonlabs/lemonade/tree/master/"

def custom_doc_links(name): return None
